import re
import uuid
from datetime import datetime
from decimal import Decimal

import transaction
from pyramid.httpexceptions import HTTPForbidden
from pyramid.view import view_config
from sqlalchemy import func

from barker.exceptions import ValidationFailure
from barker.models import (
    Inventory,
    InventoryModifier,
    Kot,
    Settlement,
    SettleOption,
    Voucher,
    Overview,
    VoucherType,
)


@view_config(request_method='PUT', route_name='voucher', renderer='json', trans=True)
def save(request):
    update_table = request.GET['u']
    json = request.json_body

    if not json['Printed'] and 'Print Kot' not in request.effective_principals:
        raise HTTPForbidden("You are not allowed to print a kot")

    if json['Printed'] and 'Print Bill' not in request.effective_principals:
        raise HTTPForbidden("You are not allowed to print bill")

    item = save_voucher(json, request.dbsession)
    if update_table:
        status = "printed" if item.is_printed else "running"
        item.status = Overview(voucher_id=None, food_table_id=item.food_table_id, status=status)
        request.dbsession.add(item.status)
    transaction.commit()
    item = request.dbsession.query(Voucher).filter(Voucher.id == item.id).first()
    return voucher_info(item)


@view_config(request_method='POST', route_name='voucher_id', renderer='json', trans=True)
def update(request):
    now = datetime.now()
    update_table = request.GET['u']
    json = request.json_body
    id = uuid.UUID(request.matchdict['id'])
    item = request.dbsession.query(Voucher).filter(Voucher.id == id).first()

    if not json['Printed'] and 'Print Kot' not in request.effective_principals:
        raise HTTPForbidden("You are not allowed to print a kot")

    if json['Printed'] and 'Print Bill' not in request.effective_principals:
        raise HTTPForbidden("You are not allowed to print bill")

    if item.is_printed and 'Edit Printed Bill' not in request.effective_principals:
        raise HTTPForbidden("You are not allowed to edit a printed bill")

    if item.is_printed and not json['Printed']:
        transaction.abort()
        raise ValidationFailure("This Bill is already printed\nCannot add a Kot to it.")

    if item.is_void:
        transaction.abort()
        raise ValidationFailure("This Bill is already void.\nReason: {0}".format(item.void_reason))

    if item.is_printed and not json['Printed']:
        transaction.abort()
        raise ValidationFailure("This Bill is already printed\nCannot add a Kot to it.")

    item.pax = json['Pax']
    item.food_table_id = json['Table']['FoodTableID']
    item.customer_id = json['Customer']['CustomerID']
    item.voucher_type = json['VoucherType']
    if not item.is_printed and json['Printed']:
        item.date = now
        type = [1, 3] if item.voucher_type in [1, 3] else [item.voucher_type]
        item.bill_id = request.dbsession.query(
            func.coalesce(func.max(Voucher.bill_id), 0) + 1
        ).filter(
            Voucher.voucher_type.in_(type)
        ).scalar()

    item.is_printed = item.is_printed or json['Printed']
    item.user_id = json['User']['UserID']
    item.last_edit_date = now
    for k in item.kots:
        for i in k.inventories:
            i.service_tax_rate, i.vat_rate = get_st_vat(
                i.service_tax_rate, i.vat_rate, item.voucher_type
            )
            i.discount = next(
                Decimal(inv['Discount']) for ko in json['Kots'] for inv in ko['Inventories'] if
                uuid.UUID(inv['InventoryID']) == i.id
            )
    for k in (k for k in json['Kots'] if k['KotID'] == '00000000-0000-0000-0000-000000000000'):
        kot = Kot(item.id, item.food_table_id, item.date, item.user_id, dbsession=request.dbsession)
        item.kots.append(kot)
        request.dbsession.add(kot)
        for index, i in enumerate(k['Inventories']):
            st, vat = get_st_vat(i['ServiceTaxRate'], i['VatRate'], json['VoucherType'])
            inv = Inventory(kot.id, uuid.UUID(i['ProductID']), i['Quantity'], i['Price'], i['Discount'],
                            i['IsHappyHour'], i['ServiceTaxID'], st, i['VatID'], vat, index)
            kot.inventories.append(inv)
            request.dbsession.add(inv)
            for m in i['Modifiers']:
                mod = InventoryModifier(None, uuid.UUID(m['ModifierID']), 0)
                inv.modifiers.append(mod)
                request.dbsession.add(mod)
    get_settlements(item, request.dbsession)
    if update_table:
        vft = request.dbsession.query(Overview).filter(Overview.voucher_id == item.id).first()
        status = "printed" if item.is_printed else "running"
        if vft is None:
            item.status = Overview(voucher_id=None, food_table_id=item.food_table_id, status=status)
            request.dbsession.add(item.status)
        else:
            vft.status = status
            vft.food_table_id = item.food_table_id
    transaction.commit()
    item = request.dbsession.query(Voucher).filter(Voucher.id == item.id).first()
    return voucher_info(item)


@view_config(request_method='POST', route_name='voucher_settle', renderer='json', permission='Settle Bill', trans=False)
def settle_voucher(request):
    update_table = request.GET['u']
    json = request.json_body
    id = uuid.UUID(request.matchdict['id'])
    item = request.dbsession.query(Voucher).filter(Voucher.id == id).first()

    allowed = [SettleOption.AMOUNT(), SettleOption.ROUND_OFF(), SettleOption.UNSETTLED()]
    for i in (j for j in json if j['Amount'] != 0):
        amount = i['Amount']
        so = i['Settled']
        so_settled = [s for s in item.settlements if s.settled == so]
        if len(so_settled) == 1:
            so_settled[0].amount = amount
        else:
            s = Settlement(item.id, so, amount)
            item.settlements.append(s)
            request.dbsession.add(s)
        allowed.append(so)
    for i in (s for s in item.settlements if s.settled not in allowed):
        item.settlements.remove(i)
        request.dbsession.delete(i)
    unsettled = get_settlements(item, request.dbsession)

    if unsettled == 0 and update_table:
        request.dbsession.query(Overview).filter(Overview.voucher_id == item.id).delete()
    transaction.commit()
    item = request.dbsession.query(Voucher).filter(Voucher.id == item.id).first()
    return voucher_info(item)


@view_config(request_method='POST', route_name='voucher_reprint', renderer='json', permission='Edit Printed Bill',
             trans=True)  # Permission error
def voucher_change(request):
    json = request.json_body
    id = uuid.UUID(request.matchdict['id'])
    item = save_voucher(json, request.dbsession)
    old = request.dbsession.query(Voucher).filter(Voucher.id == id).first()
    old.void = True
    old.void_reason = "Bill Discounted / Changed. New Bill ID is {0}".format(item.full_bill_id)
    # TODO: Set the Void Settlement

    if old.status is None:
        item.status = Overview(voucher_id=None, food_table_id=item.food_table_id, status="printed")
        request.dbsession.add(item.status)
    else:
        request.dbsession.query(
            Overview
        ).filter(
            Overview.voucher_id == old.id
        ).update({Overview.voucher_id: item.id})
    transaction.commit()
    item = request.dbsession.query(Voucher).filter(Voucher.id == item.id).first()
    return voucher_info(item)


@view_config(request_method='POST', route_name='voucher_void', renderer='json', permission='Void Bill', trans=True)
def void_voucher(request):
    update_table = request.GET['u']
    json = request.json_body
    id = uuid.UUID(request.matchdict['id'])
    item = request.dbsession.query(Voucher).filter(Voucher.id == id).first()

    item.void = True
    item.void_reason = json['Reason']

    amount = round(-1 * sum(i.amount for k in item.kots for i in k.inventories), 5) * -1
    s = Settlement(item.id, SettleOption.VOID(), amount)
    item.settlements.append(s)
    request.dbsession.add(s)
    allowed = [SettleOption.AMOUNT(), SettleOption.ROUND_OFF(), SettleOption.UNSETTLED(), SettleOption.VOID()]
    for i in (s for s in item.settlements if s.settled not in allowed):
        item.settlements.remove(i)
        request.dbsession.delete(i)
    get_settlements(item, request.dbsession)

    if update_table:
        request.dbsession.delete(item.status)
    transaction.commit()
    item = request.dbsession.query(Voucher).filter(Voucher.id == item.id).first()
    return voucher_info(item)


@view_config(request_method='POST', route_name='voucher_split', renderer='json', permission='Split Bill', trans=True)
def split_voucher(request):
    json = request.json_body
    id = uuid.UUID(request.matchdict['id'])
    item = request.dbsession.query(Voucher).filter(Voucher.id == id).first()
    item.void = True
    item.void_reason = 'Bill Split'
    # TODO: Set the Void Settlement

    new_one = save_voucher(json['One'], request.dbsession)
    new_two = save_voucher(json['Two'], request.dbsession)

    status = "printed" if item.is_printed else "running"
    new_one.status = Overview(voucher_id=None, food_table_id=new_one.food_table_id, status=status)

    request.dbsession.add(new_one.status)
    request.dbsession.query(
        Overview
    ).filter(
        Overview.voucher_id == item.id
    ).update({Overview.voucher_id: new_two.id})

    transaction.commit()


@view_config(request_method='GET', route_name='voucher_id', renderer='json', permission='Authenticated')
def show_id(request):
    id = uuid.UUID(request.matchdict['id'])
    item = request.dbsession.query(Voucher).filter(Voucher.id == id).first()
    return voucher_info(item)


@view_config(request_method='GET', route_name='voucher', request_param='b', renderer='json', permission='Authenticated')
def show(request):
    bill_id = request.GET['b']
    item = request.dbsession.query(Voucher)
    if re.compile('^\d{2,}-\d{4}$').match(bill_id):
        item = item.filter(
            Voucher.bill_id == int(bill_id.replace('-', '')),
            Voucher.voucher_type.in_([1, 3])
        )
    elif re.compile('^NC-\d+$').match(bill_id):
        item = item.filter(
            Voucher.bill_id == int(bill_id.replace('NC-', '')),
            Voucher.voucher_type == 2
        )
    elif re.compile('^ST-\d+$').match(bill_id):
        item = item.filter(
            Voucher.bill_id == int(bill_id.replace('ST-', '')),
            Voucher.voucher_type == 4
        )
    item = item.first()
    if item is None:
        return {}
    return voucher_info(item)


def voucher_info(item):
    return {
        'VoucherID': item.id,
        'Date': item.date.strftime('%d-%b-%Y %H:%M:%S'),
        'Pax': item.pax,
        'User': {'UserID': item.user_id, 'Name': item.user.name},
        'CreationDate': item.creation_date.strftime('%d-%b-%Y %H:%M:%S'),
        'LastEditDate': item.last_edit_date.strftime('%d-%b-%Y %H:%M:%S'),
        'BillID': item.bill_id,
        'Table': {'FoodTableID': item.food_table_id, 'Name': item.food_table.name},
        'Customer': {'CustomerID': item.customer_id, 'Name': item.customer.name},
        'Settlements': [],
        'Narration': item.narration,
        'Void': item.is_void,
        'VoidReason': item.void_reason,
        'Printed': item.is_printed,
        'VoucherType': item.voucher_type,
        'KotID': item.kot_id,
        'Kots': [
            {
                'KotID': k.id, 'Code': k.code, 'Date': k.date.strftime('%d-%b-%Y %H:%M:%S'),
                'User': {'UserID': k.user_id, 'Name': k.user.name},
                'Inventories': [
                    {
                        'InventoryID': i.id,
                        'SortOrder': i.sort_order,
                        'Product':
                            {
                                'ProductID': i.product_id,
                                'Name': i.product.name,
                                'Units': i.product.units,
                                'ProductGroup':
                                    {
                                        'ProductGroupID': i.product.product_group_id,
                                        'Name': i.product.product_group.name,
                                        'GroupType': i.product.product_group.group_type,
                                        'DiscountLimit': i.product.product_group.discount_limit
                                    }
                            },
                        'Quantity': i.quantity,
                        'Price': i.price,
                        'IsHappyHour': i.is_happy_hour,
                        'ServiceTaxRate': i.service_tax_rate,
                        'VatRate': i.vat_rate,
                        'ServiceTax': {'TaxID': i.service_tax_id, 'Name': i.service_tax.name},
                        'Vat': {'TaxID': i.vat_id, 'Name': i.vat.name},
                        'Discount': i.discount,
                        'InventoryModifier': [
                            {
                                'Modifier': {
                                    'ModifierID': m.modifier.id,
                                    'Name': m.modifier.name,
                                    'ShowInBill': m.modifier.show_in_bill
                                },
                                'Price': m.price
                            } for m in i.modifiers]
                    } for i in k.inventories]
            } for k in item.kots],
        'Reprints': []
    }


def get_st_vat(st, vat, voucher_type):
    if voucher_type == VoucherType.STAFF.value:
        return 0, 0
    elif voucher_type == VoucherType.NO_CHARGE.value:
        return 0, 0
    else:  # voucher_type in [REGULAR_BILL, TAKE_AWAY]:
        return st, vat


def get_settlements(voucher, dbsession):
    amount = round(-1 * sum(sum(i.amount for i in k.inventories) for k in voucher.kots), 5)
    so_amount = [s for s in voucher.settlements if s.settled == SettleOption.AMOUNT()]
    if len(so_amount) == 1:
        so_amount[0].amount = amount
    else:
        s = Settlement(voucher.id, SettleOption.AMOUNT(), amount)
        voucher.settlements.append(s)
        dbsession.add(s)

    round_off = round(amount) - amount
    so_round_off = [s for s in voucher.settlements if s.settled == SettleOption.ROUND_OFF()]
    if len(so_round_off) == 1 and round_off != 0:
        so_round_off[0].amount = round_off
    elif len(so_round_off) == 1 and round_off == 0:
        voucher.settlements.remove(so_round_off[0])
        dbsession.delete(so_round_off[0])
    elif len(so_round_off) == 0 and round_off != 0:
        s = Settlement(voucher.id, SettleOption.ROUND_OFF(), round_off)
        voucher.settlements.append(s)
        dbsession.add(s)

    unsettled = sum(s.amount for s in voucher.settlements if s.settled != SettleOption.UNSETTLED())
    so_unsettled = [s for s in voucher.settlements if s.settled == SettleOption.UNSETTLED()]
    if len(so_unsettled) == 1 and unsettled != 0:
        so_unsettled[0].amount = unsettled
    elif len(so_unsettled) == 1 and unsettled == 0:
        voucher.settlements.remove(so_unsettled[0])
        dbsession.delete(so_unsettled[0])
    elif len(so_unsettled) == 0 and unsettled != 0:
        s = Settlement(voucher.id, SettleOption.UNSETTLED(), unsettled)
        voucher.settlements.append(s)
        dbsession.add(s)
    return unsettled


def save_voucher(json, dbsession):
    item = Voucher(json['Pax'], json['Table']['FoodTableID'], json['Customer']['CustomerID'], json['Printed'],
                   json['VoucherType'], json['User']['UserID'], dbsession)
    dbsession.add(item)
    for k in json['Kots']:
        kot = Kot(item.id, item.food_table_id, item.date, item.user_id, dbsession=dbsession)
        item.kots.append(kot)
        dbsession.add(kot)
        for index, i in enumerate(k['Inventories']):
            st, vat = get_st_vat(i['ServiceTaxRate'], i['VatRate'], json['VoucherType'])
            inv = Inventory(kot.id, uuid.UUID(i['ProductID']), i['Quantity'], i['Price'], i['Discount'],
                            i['IsHappyHour'], i['ServiceTaxID'], st, i['VatID'], vat, index)
            kot.inventories.append(inv)
            dbsession.add(inv)
            for m in i['Modifiers']:
                mod = InventoryModifier(None, uuid.UUID(m['ModifierID']), 0)
                inv.modifiers.append(mod)
                dbsession.add(mod)
    get_settlements(item, dbsession)
    return item
